# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/loader.base.ipynb.

# %% ../../nbs/loader.base.ipynb 3
from __future__ import print_function, division, annotations
from ..imports import *

# %% auto 0
__all__ = ['BaseDataLoader']

# %% ../../nbs/loader.base.ipynb 4
class BaseDataLoader:
    """Dataloader Interface"""
    
    def __init__(
        self, 
        dataset, 
        batch_size: int = 1,  # batch size
        shuffle: bool = False,  # if true, dataloader shuffles before sampling each batch
        num_workers: int = 0,  # how many subprocesses to use for data loading.
        drop_last: bool = False,
        **kwargs
    ):
        pass

    def __len__(self):
        raise NotImplementedError
    
    def __next__(self):
        raise NotImplementedError
    
    def __iter__(self):
        raise NotImplementedError



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/loader.torch.ipynb.

# %% ../../nbs/loader.torch.ipynb 2
from __future__ import print_function, division, annotations
from ..imports import *
from . import BaseDataLoader
from ..datasets import Dataset, ArrayDataset, JAXDataset
from ..utils import check_pytorch_installed
from ..tests import *
from jax.tree_util import tree_map
import warnings


# %% auto 0
__all__ = ['to_torch_dataset', 'DataLoaderPytorch']

# %% ../../nbs/loader.torch.ipynb 5
# adapted from https://jax.readthedocs.io/en/latest/notebooks/Neural_Network_and_Data_Loading.html
def _numpy_collate(batch):
  return tree_map(np.asarray, torch_data.default_collate(batch))

# %% ../../nbs/loader.torch.ipynb 6
@dispatch
def to_torch_dataset(dataset: JAXDataset) -> torch_data.Dataset:
    class DatasetPytorch(torch_data.Dataset):
        def __init__(self, dataset: Dataset): self.dataset = dataset
        def __len__(self): return len(self.dataset)
        def __getitem__(self, idx): return self.dataset[idx]
    
    return DatasetPytorch(dataset)

@dispatch
def to_torch_dataset(dataset: TorchDataset):
    return dataset

@dispatch
def to_torch_dataset(dataset: HFDataset):
    return dataset.with_format("numpy")

# %% ../../nbs/loader.torch.ipynb 7
class DataLoaderPytorch(BaseDataLoader):
    """Pytorch Dataloader"""
    
    @typecheck
    def __init__(
        self, 
        dataset: Union[JAXDataset, TorchDataset, HFDataset],
        batch_size: int = 1,  # Batch size
        shuffle: bool = False,  # If true, dataloader shuffles before sampling each batch
        drop_last: bool = False, # Drop last batch or not
        **kwargs
    ):
        super().__init__(dataset, batch_size, shuffle, drop_last)
        check_pytorch_installed()
        from torch.utils.data import BatchSampler, RandomSampler, SequentialSampler

        if 'sampler' in kwargs:
            warnings.warn("`sampler` is currently not supported. We will ignore it and use `shuffle` instead.")
            del kwargs['sampler']

        dataset = to_torch_dataset(dataset)
        sampler = RandomSampler(dataset) if shuffle else SequentialSampler(dataset)
        batch_sampler = BatchSampler(sampler, batch_size=batch_size, drop_last=drop_last)

        self.dataloader = torch_data.DataLoader(
            dataset, 
            batch_sampler=batch_sampler,
            # batch_size=batch_size, 
            # shuffle=shuffle, 
            # drop_last=drop_last,
            collate_fn=_numpy_collate,
            **kwargs
        )

    def __len__(self):
        return len(self.dataloader)

    def __next__(self):
        return next(self.dataloader)

    def __iter__(self):
        return self.dataloader.__iter__()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/loader.tf.ipynb.

# %% ../../nbs/loader.tf.ipynb 2
from __future__ import print_function, division, annotations
from ..imports import *
from . import BaseDataLoader
from ..datasets import Dataset, ArrayDataset, JAXDataset
from ..utils import check_tf_installed, get_config
from ..tests import *
from jax.tree_util import tree_map

# %% auto 0
__all__ = ['to_tf_dataset', 'DataLoaderTensorflow']

# %% ../../nbs/loader.tf.ipynb 4
@dispatch
def to_tf_dataset(dataset: JAXDataset) -> tf.data.Dataset:
    return tf.data.Dataset.from_tensor_slices(dataset[:])

@dispatch
def to_tf_dataset(dataset: TFDataset) -> tf.data.Dataset:
    return dataset

@dispatch
def to_tf_dataset(dataset: HFDataset) -> tf.data.Dataset:
    return dataset.to_tf_dataset()

# %% ../../nbs/loader.tf.ipynb 5
class DataLoaderTensorflow(BaseDataLoader):
    """Tensorflow Dataloader"""
    
    @typecheck
    def __init__(
        self, 
        dataset: Union[JAXDataset, TFDataset, HFDataset],
        batch_size: int = 1,  # Batch size
        shuffle: bool = False,  # If true, dataloader shuffles before sampling each batch
        drop_last: bool = False, # Drop last batch or not
        **kwargs
    ):
        super().__init__(dataset, batch_size, shuffle, drop_last)
        check_tf_installed()
        # Convert to tf dataset
        ds = to_tf_dataset(dataset)
        ds = ds.shuffle(buffer_size=len(dataset), seed=get_config().global_seed) if shuffle else ds
        ds = ds.batch(batch_size, drop_remainder=drop_last)
        ds = ds.prefetch(tf.data.AUTOTUNE)
        self.dataloader = ds

    def __len__(self):
        return len(self.dataloader)

    def __next__(self):
        return next(self.dataloader)

    def __iter__(self):
        return self.dataloader.as_numpy_iterator()



from __future__ import annotations
import numpy as np
from typing import (
    List,
    Tuple,
    Dict,
    Any,
    Optional,
    Iterable,
    Sequence,
    Iterator,
    Literal,
    Union,
    Annotated,
)
import jax
from jax import vmap, grad, jit, numpy as jnp, random as jrand
from abc import ABC
from dataclasses import dataclass
from plum import dispatch
from beartype.vale import Is
from beartype.door import is_bearable
from beartype import beartype as typecheck

try:
    import torch.utils.data as torch_data
    import torch

    TorchDataset = torch_data.Dataset
except ModuleNotFoundError:
    torch_data = None
    torch = None
    TorchDataset = Annotated[None, Is[lambda _: torch_data is not None]]

try:
    import datasets as hf_datasets

    HFDataset = Annotated[
        Union[
            hf_datasets.Dataset,
            hf_datasets.DatasetDict,
            hf_datasets.IterableDatasetDict,
            hf_datasets.IterableDataset,
        ],
        Is[lambda _: hf_datasets is not None],
    ]
except ModuleNotFoundError:
    hf_datasets = None
    HFDataset = Annotated[None, Is[lambda _: hf_datasets is not None]]

try:
    import tensorflow as tf
    import tensorflow_datasets as tfds

    # TFDataset = Annotated[
    #     tf.data.Dataset,
    #     Is[lambda _: tf is not None],
    # ]
    TFDataset = tf.data.Dataset
except ModuleNotFoundError:
    tf = None
    tfds = None
    TFDataset = Annotated[None, Is[lambda _: tf is not None]]

try:
    import haiku as hk
except ModuleNotFoundError:
    hk = None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/utils.ipynb.

# %% ../nbs/utils.ipynb 4
from __future__ import print_function, division, annotations
from .imports import *
import jax_dataloader as jdl
import collections

# %% auto 0
__all__ = ['Config', 'get_config', 'check_pytorch_installed', 'has_pytorch_tensor', 'check_hf_installed', 'check_tf_installed',
           'asnumpy']

# %% ../nbs/utils.ipynb 6
@dataclass
class Config:
    """Global configuration for the library"""
    rng_reserve_size: int
    global_seed: int

    @classmethod
    def default(cls) -> Config:
        return cls(rng_reserve_size=1, global_seed=42)

# %% ../nbs/utils.ipynb 7
main_config = Config.default()

# %% ../nbs/utils.ipynb 8
def get_config() -> Config:
    return main_config

# %% ../nbs/utils.ipynb 10
def check_pytorch_installed():
    if torch_data is None:
        raise ModuleNotFoundError("`pytorch` library needs to be installed. "
            "Try `pip install torch`. Please refer to pytorch documentation for details: "
            "https://pytorch.org/get-started/.")


# %% ../nbs/utils.ipynb 12
def has_pytorch_tensor(batch) -> bool:
    if isinstance(batch[0], torch.Tensor):
        return True
    elif isinstance(batch[0], (tuple, list)):
        transposed = zip(*batch)
        return any([has_pytorch_tensor(samples) for samples in transposed])
    else:
        return False

# %% ../nbs/utils.ipynb 13
def check_hf_installed():
    if hf_datasets is None:
        raise ModuleNotFoundError("`datasets` library needs to be installed. "
            "Try `pip install datasets`. Please refer to huggingface documentation for details: "
            "https://huggingface.co/docs/datasets/installation.html.")

# %% ../nbs/utils.ipynb 15
def check_tf_installed():
    if tf is None:
        raise ModuleNotFoundError("`tensorflow` library needs to be installed. "
            "Try `pip install tensorflow`. Please refer to tensorflow documentation for details: "
            "https://www.tensorflow.org/install/pip.")

# %% ../nbs/utils.ipynb 18
def asnumpy(x) -> np.ndarray:
    if isinstance(x, np.ndarray):
        return x
    elif isinstance(x, jnp.ndarray):
        return x.__array__()
    elif torch_data and isinstance(x, torch.Tensor):
        return x.detach().cpu().numpy()
    elif tf and isinstance(x, tf.Tensor):
        return x.numpy()
    elif isinstance(x, (tuple, list)):
        return map(asnumpy, x)
    else:
        raise ValueError(f"Unknown type: {type(x)}")



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/jax-dataloader',
                'doc_host': 'https://birkhoffg.github.io',
                'git_url': 'https://github.com/birkhoffg/jax-dataloader',
                'lib_path': 'jax_dataloader'},
  'syms': { 'jax_dataloader.core': { 'jax_dataloader.core.DataLoader': ('core.html#dataloader', 'jax_dataloader/core.py'),
                                     'jax_dataloader.core.DataLoader.__init__': ('core.html#dataloader.__init__', 'jax_dataloader/core.py'),
                                     'jax_dataloader.core.DataLoader.__iter__': ('core.html#dataloader.__iter__', 'jax_dataloader/core.py'),
                                     'jax_dataloader.core.DataLoader.__len__': ('core.html#dataloader.__len__', 'jax_dataloader/core.py'),
                                     'jax_dataloader.core.DataLoader.__next__': ('core.html#dataloader.__next__', 'jax_dataloader/core.py'),
                                     'jax_dataloader.core.DataloaderBackends': ('core.html#dataloaderbackends', 'jax_dataloader/core.py'),
                                     'jax_dataloader.core.DataloaderBackends.__getitem__': ( 'core.html#dataloaderbackends.__getitem__',
                                                                                             'jax_dataloader/core.py'),
                                     'jax_dataloader.core.DataloaderBackends.supported': ( 'core.html#dataloaderbackends.supported',
                                                                                           'jax_dataloader/core.py'),
                                     'jax_dataloader.core._check_backend_compatibility': ( 'core.html#_check_backend_compatibility',
                                                                                           'jax_dataloader/core.py'),
                                     'jax_dataloader.core._dispatch_dataloader': ( 'core.html#_dispatch_dataloader',
                                                                                   'jax_dataloader/core.py'),
                                     'jax_dataloader.core._get_backends': ('core.html#_get_backends', 'jax_dataloader/core.py'),
                                     'jax_dataloader.core.get_backend_compatibilities': ( 'core.html#get_backend_compatibilities',
                                                                                          'jax_dataloader/core.py')},
            'jax_dataloader.datasets': { 'jax_dataloader.datasets.ArrayDataset': ( 'dataset.html#arraydataset',
                                                                                   'jax_dataloader/datasets.py'),
                                         'jax_dataloader.datasets.ArrayDataset.__getitem__': ( 'dataset.html#arraydataset.__getitem__',
                                                                                               'jax_dataloader/datasets.py'),
                                         'jax_dataloader.datasets.ArrayDataset.__init__': ( 'dataset.html#arraydataset.__init__',
                                                                                            'jax_dataloader/datasets.py'),
                                         'jax_dataloader.datasets.ArrayDataset.__len__': ( 'dataset.html#arraydataset.__len__',
                                                                                           'jax_dataloader/datasets.py'),
                                         'jax_dataloader.datasets.ArrayDataset.asnumpy': ( 'dataset.html#arraydataset.asnumpy',
                                                                                           'jax_dataloader/datasets.py'),
                                         'jax_dataloader.datasets.Dataset': ('dataset.html#dataset', 'jax_dataloader/datasets.py'),
                                         'jax_dataloader.datasets.Dataset.__getitem__': ( 'dataset.html#dataset.__getitem__',
                                                                                          'jax_dataloader/datasets.py'),
                                         'jax_dataloader.datasets.Dataset.__len__': ( 'dataset.html#dataset.__len__',
                                                                                      'jax_dataloader/datasets.py')},
            'jax_dataloader.experimental.multi_processing': { 'jax_dataloader.experimental.multi_processing.EpochIterator': ( 'experimental/mp.html#epochiterator',
                                                                                                                              'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.EpochIterator.__del__': ( 'experimental/mp.html#epochiterator.__del__',
                                                                                                                                      'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.EpochIterator.__init__': ( 'experimental/mp.html#epochiterator.__init__',
                                                                                                                                       'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.EpochIterator.__iter__': ( 'experimental/mp.html#epochiterator.__iter__',
                                                                                                                                       'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.EpochIterator.__next__': ( 'experimental/mp.html#epochiterator.__next__',
                                                                                                                                       'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.EpochIterator.close': ( 'experimental/mp.html#epochiterator.close',
                                                                                                                                    'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.EpochIterator.get_data': ( 'experimental/mp.html#epochiterator.get_data',
                                                                                                                                       'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.EpochIterator.run': ( 'experimental/mp.html#epochiterator.run',
                                                                                                                                  'jax_dataloader/experimental/multi_processing.py'),
                                                              'jax_dataloader.experimental.multi_processing.chunk': ( 'experimental/mp.html#chunk',
                                                                                                                      'jax_dataloader/experimental/multi_processing.py')},
            'jax_dataloader.imports': {},
            'jax_dataloader.loaders.base': { 'jax_dataloader.loaders.base.BaseDataLoader': ( 'loader.base.html#basedataloader',
                                                                                             'jax_dataloader/loaders/base.py'),
                                             'jax_dataloader.loaders.base.BaseDataLoader.__init__': ( 'loader.base.html#basedataloader.__init__',
                                                                                                      'jax_dataloader/loaders/base.py'),
                                             'jax_dataloader.loaders.base.BaseDataLoader.__iter__': ( 'loader.base.html#basedataloader.__iter__',
                                                                                                      'jax_dataloader/loaders/base.py'),
                                             'jax_dataloader.loaders.base.BaseDataLoader.__len__': ( 'loader.base.html#basedataloader.__len__',
                                                                                                     'jax_dataloader/loaders/base.py'),
                                             'jax_dataloader.loaders.base.BaseDataLoader.__next__': ( 'loader.base.html#basedataloader.__next__',
                                                                                                      'jax_dataloader/loaders/base.py')},
            'jax_dataloader.loaders.jax': { 'jax_dataloader.loaders.jax.DataLoaderJAX': ( 'loader.jax.html#dataloaderjax',
                                                                                          'jax_dataloader/loaders/jax.py'),
                                            'jax_dataloader.loaders.jax.DataLoaderJAX.__init__': ( 'loader.jax.html#dataloaderjax.__init__',
                                                                                                   'jax_dataloader/loaders/jax.py'),
                                            'jax_dataloader.loaders.jax.DataLoaderJAX.__iter__': ( 'loader.jax.html#dataloaderjax.__iter__',
                                                                                                   'jax_dataloader/loaders/jax.py'),
                                            'jax_dataloader.loaders.jax.DataLoaderJAX.__len__': ( 'loader.jax.html#dataloaderjax.__len__',
                                                                                                  'jax_dataloader/loaders/jax.py'),
                                            'jax_dataloader.loaders.jax.DataLoaderJAX.next_key': ( 'loader.jax.html#dataloaderjax.next_key',
                                                                                                   'jax_dataloader/loaders/jax.py'),
                                            'jax_dataloader.loaders.jax.EpochIterator': ( 'loader.jax.html#epochiterator',
                                                                                          'jax_dataloader/loaders/jax.py'),
                                            'jax_dataloader.loaders.jax.to_jax_dataset': ( 'loader.jax.html#to_jax_dataset',
                                                                                           'jax_dataloader/loaders/jax.py')},
            'jax_dataloader.loaders.tensorflow': { 'jax_dataloader.loaders.tensorflow.DataLoaderTensorflow': ( 'loader.tf.html#dataloadertensorflow',
                                                                                                               'jax_dataloader/loaders/tensorflow.py'),
                                                   'jax_dataloader.loaders.tensorflow.DataLoaderTensorflow.__init__': ( 'loader.tf.html#dataloadertensorflow.__init__',
                                                                                                                        'jax_dataloader/loaders/tensorflow.py'),
                                                   'jax_dataloader.loaders.tensorflow.DataLoaderTensorflow.__iter__': ( 'loader.tf.html#dataloadertensorflow.__iter__',
                                                                                                                        'jax_dataloader/loaders/tensorflow.py'),
                                                   'jax_dataloader.loaders.tensorflow.DataLoaderTensorflow.__len__': ( 'loader.tf.html#dataloadertensorflow.__len__',
                                                                                                                       'jax_dataloader/loaders/tensorflow.py'),
                                                   'jax_dataloader.loaders.tensorflow.DataLoaderTensorflow.__next__': ( 'loader.tf.html#dataloadertensorflow.__next__',
                                                                                                                        'jax_dataloader/loaders/tensorflow.py'),
                                                   'jax_dataloader.loaders.tensorflow.to_tf_dataset': ( 'loader.tf.html#to_tf_dataset',
                                                                                                        'jax_dataloader/loaders/tensorflow.py')},
            'jax_dataloader.loaders.torch': { 'jax_dataloader.loaders.torch.DataLoaderPytorch': ( 'loader.torch.html#dataloaderpytorch',
                                                                                                  'jax_dataloader/loaders/torch.py'),
                                              'jax_dataloader.loaders.torch.DataLoaderPytorch.__init__': ( 'loader.torch.html#dataloaderpytorch.__init__',
                                                                                                           'jax_dataloader/loaders/torch.py'),
                                              'jax_dataloader.loaders.torch.DataLoaderPytorch.__iter__': ( 'loader.torch.html#dataloaderpytorch.__iter__',
                                                                                                           'jax_dataloader/loaders/torch.py'),
                                              'jax_dataloader.loaders.torch.DataLoaderPytorch.__len__': ( 'loader.torch.html#dataloaderpytorch.__len__',
                                                                                                          'jax_dataloader/loaders/torch.py'),
                                              'jax_dataloader.loaders.torch.DataLoaderPytorch.__next__': ( 'loader.torch.html#dataloaderpytorch.__next__',
                                                                                                           'jax_dataloader/loaders/torch.py'),
                                              'jax_dataloader.loaders.torch._numpy_collate': ( 'loader.torch.html#_numpy_collate',
                                                                                               'jax_dataloader/loaders/torch.py'),
                                              'jax_dataloader.loaders.torch.to_torch_dataset': ( 'loader.torch.html#to_torch_dataset',
                                                                                                 'jax_dataloader/loaders/torch.py')},
            'jax_dataloader.tests': { 'jax_dataloader.tests.get_batch': ('tests.html#get_batch', 'jax_dataloader/tests.py'),
                                      'jax_dataloader.tests.test_dataloader': ('tests.html#test_dataloader', 'jax_dataloader/tests.py'),
                                      'jax_dataloader.tests.test_no_shuffle': ('tests.html#test_no_shuffle', 'jax_dataloader/tests.py'),
                                      'jax_dataloader.tests.test_no_shuffle_drop_last': ( 'tests.html#test_no_shuffle_drop_last',
                                                                                          'jax_dataloader/tests.py'),
                                      'jax_dataloader.tests.test_shuffle': ('tests.html#test_shuffle', 'jax_dataloader/tests.py'),
                                      'jax_dataloader.tests.test_shuffle_drop_last': ( 'tests.html#test_shuffle_drop_last',
                                                                                       'jax_dataloader/tests.py')},
            'jax_dataloader.utils': { 'jax_dataloader.utils.Config': ('utils.html#config', 'jax_dataloader/utils.py'),
                                      'jax_dataloader.utils.Config.default': ('utils.html#config.default', 'jax_dataloader/utils.py'),
                                      'jax_dataloader.utils.asnumpy': ('utils.html#asnumpy', 'jax_dataloader/utils.py'),
                                      'jax_dataloader.utils.check_hf_installed': ( 'utils.html#check_hf_installed',
                                                                                   'jax_dataloader/utils.py'),
                                      'jax_dataloader.utils.check_pytorch_installed': ( 'utils.html#check_pytorch_installed',
                                                                                        'jax_dataloader/utils.py'),
                                      'jax_dataloader.utils.check_tf_installed': ( 'utils.html#check_tf_installed',
                                                                                   'jax_dataloader/utils.py'),
                                      'jax_dataloader.utils.get_config': ('utils.html#get_config', 'jax_dataloader/utils.py'),
                                      'jax_dataloader.utils.has_pytorch_tensor': ( 'utils.html#has_pytorch_tensor',
                                                                                   'jax_dataloader/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/core.ipynb.

# %% ../nbs/core.ipynb 3
from __future__ import print_function, division, annotations
from .imports import *
from .utils import *
from .datasets import *
from .loaders import *

# %% auto 0
__all__ = ['SUPPORTED_DATASETS', 'DataloaderBackends', 'get_backend_compatibilities', 'DataLoader']

# %% ../nbs/core.ipynb 4
SUPPORTED_DATASETS = [
    JAXDataset,
    TorchDataset,
    TFDataset,
    HFDataset
]

# %% ../nbs/core.ipynb 5
@dataclass(frozen=True)
class DataloaderBackends:
    jax = DataLoaderJAX
    pytorch: BaseDataLoader = DataLoaderPytorch
    tensorflow: BaseDataLoader = DataLoaderTensorflow
    merlin: BaseDataLoader = None

    __all__ = dict(
        jax=jax, pytorch=pytorch, tensorflow=tensorflow, merlin=merlin
    )

    def __getitem__(self, key):
        return self.__all__[key]

    @property
    def supported(self) -> List[str]:
        return [
            backend for backend, dl_cls in self.__all__.items() if dl_cls is not None
        ]

# %% ../nbs/core.ipynb 6
def _get_backends() -> List[str]:
    """Return list of supported dataloader backends"""
    return DataloaderBackends().__all__.keys()


def _dispatch_dataloader(
    backend: str # dataloader backend
) -> BaseDataLoader:
    """Return Dataloader class based on given `backend`"""
    backends = DataloaderBackends()
    if not backend in backends.supported:
        raise ValueError(f"backend=`{backend}` is either an invalid backend or not supported yet. "
            f"Should be one of {backends.supported}.")
    
    dl_cls = backends[backend]
    return dl_cls

# %% ../nbs/core.ipynb 7
def _check_backend_compatibility(ds, backend: str):
    return DataLoader(ds, backend=backend)

# %% ../nbs/core.ipynb 8
def get_backend_compatibilities() -> dict[str, list[type]]: # { backend: [supported datasets] }
    """Return list of supported dataloader backends for each dataset type"""

    ds = {
        JAXDataset: ArrayDataset(np.array([1,2,3])),
        TorchDataset: torch_data.Dataset(),
        TFDataset: tf.data.Dataset.from_tensor_slices(np.array([1,2,3])),
        HFDataset: hf_datasets.Dataset.from_dict({'a': [1,2,3]})
    }
    assert len(ds) == len(SUPPORTED_DATASETS)
    backends = {b: [] for b in _get_backends()}
    for b in _get_backends():
        for name, dataset in ds.items():
            try:
                _check_backend_compatibility(dataset, b)
                backends[b].append(name)
            except:
                pass

    return backends

# %% ../nbs/core.ipynb 9
class DataLoader:
    """Main Dataloader class to load Numpy data batches"""

    def __init__(
        self,
        dataset, # Dataset from which to load the data
        backend: Literal['jax', 'pytorch', 'tensorflow'], # Dataloader backend to load the dataset
        batch_size: int = 1,  # How many samples per batch to load
        shuffle: bool = False,  # If true, dataloader reshuffles every epoch
        drop_last: bool = False, # If true, drop the last incomplete batch
        **kwargs
    ):
        dl_cls = _dispatch_dataloader(backend)
        self.dataloader = dl_cls(
            dataset=dataset, 
            batch_size=batch_size, 
            shuffle=shuffle, 
            drop_last=drop_last,
            **kwargs
        )

    def __len__(self):
        return len(self.dataloader)

    def __next__(self):
        return next(self.dataloader)

    def __iter__(self):
        return iter(self.dataloader)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/dataset.ipynb.

# %% ../nbs/dataset.ipynb 3
from __future__ import print_function, division, annotations
from .imports import *
from .utils import asnumpy

# %% auto 0
__all__ = ['JAXDataset', 'Dataset', 'ArrayDataset']

# %% ../nbs/dataset.ipynb 4
class Dataset:
    """A pytorch-like Dataset class."""

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

# %% ../nbs/dataset.ipynb 5
class ArrayDataset(Dataset):
    """Dataset wrapping numpy arrays."""

    def __init__(
        self, 
        *arrays: jax.Array, # Numpy array with same first dimension
        asnumpy: bool = True, # Store arrays as numpy arrays if True; otherwise store as array type of *arrays
    ):
        assert all(arrays[0].shape[0] == arr.shape[0] for arr in arrays), \
            "All arrays must have the same dimension."
        self.arrays = tuple(arrays)
        if asnumpy:
            self.asnumpy()            
    
    def asnumpy(self):
        """Convert all arrays to numpy arrays."""
        self.arrays = tuple(asnumpy(arr) for arr in self.arrays)

    def __len__(self):
        return self.arrays[0].shape[0]

    def __getitem__(self, index):
        return jax.tree_util.tree_map(lambda x: x[index], self.arrays)

# %% ../nbs/dataset.ipynb 14
JAXDataset = Dataset



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/tests.ipynb.

# %% ../nbs/tests.ipynb 2
from __future__ import print_function, division, annotations
from .imports import *
from .datasets import ArrayDataset

# %% auto 0
__all__ = ['test_dataloader']

# %% ../nbs/tests.ipynb 3
def get_batch(batch):
    if isinstance(batch, dict):
        return batch['feats'], batch['labels']
    else:
        return batch

# %% ../nbs/tests.ipynb 4
def test_no_shuffle(cls, ds, batch_size: int, feats, labels):
    dl = cls(ds, batch_size=batch_size, shuffle=False)
    assert len(dl) == len(feats) // batch_size + 1
    for _ in range(2):
        X_list, Y_list = [], []
        for batch in dl:
            x, y = get_batch(batch)
            X_list.append(x)
            Y_list.append(y)
        _X, _Y = map(jnp.concatenate, (X_list, Y_list))
        assert jnp.array_equal(_X, feats)
        assert jnp.array_equal(_Y, labels)

# %% ../nbs/tests.ipynb 5
def test_no_shuffle_drop_last(cls, ds, batch_size: int, feats, labels):
    dl = cls(ds, batch_size=batch_size, shuffle=False, drop_last=True)
    assert len(dl) == len(feats) // batch_size
    for _ in range(2):
        X_list, Y_list = [], []
        for batch in dl:
            x, y = get_batch(batch)
            X_list.append(x)
            Y_list.append(y)
        _X, _Y = map(jnp.concatenate, (X_list, Y_list))
        last_idx = len(X_list) * batch_size
        assert jnp.array_equal(_X, feats[: last_idx])
        assert jnp.array_equal(_Y, labels[: last_idx])

# %% ../nbs/tests.ipynb 6
def test_shuffle(cls, ds, batch_size: int, feats, labels):
    dl = cls(ds, batch_size=batch_size, shuffle=True, drop_last=False)
    last_X, last_Y = jnp.array([]), jnp.array([])
    assert len(dl) == len(feats) // batch_size + 1
    for _ in range(2):
        X_list, Y_list = [], []
        for batch in dl:
            x, y = get_batch(batch)
            assert jnp.array_equal(x[:, :1], y)
            X_list.append(x)
            Y_list.append(y)
        _X, _Y = map(jnp.concatenate, (X_list, Y_list))
        assert not jnp.array_equal(_X, feats)
        assert not jnp.array_equal(_Y, labels)
        assert jnp.sum(_X) == jnp.sum(feats), \
            f"jnp.sum(_X)={jnp.sum(_X)}, jnp.sum(feats)={jnp.sum(feats)}"
        assert not jnp.array_equal(_X, last_X)
        assert not jnp.array_equal(_Y, last_Y)
        last_X, last_Y = _X, _Y

# %% ../nbs/tests.ipynb 7
def test_shuffle_drop_last(cls, ds, batch_size: int, feats, labels):
    dl = cls(ds, batch_size=batch_size, shuffle=True, drop_last=True)
    assert len(dl) == len(feats) // batch_size
    for _ in range(2):
        X_list, Y_list = [], []
        for batch in dl:
            x, y = get_batch(batch)
            assert jnp.array_equal(x[:, :1], y)
            X_list.append(x)
            Y_list.append(y)
        _X, _Y = map(jnp.concatenate, (X_list, Y_list))
        assert not jnp.array_equal(_X, feats)
        assert not jnp.array_equal(_Y, labels)
        assert len(_X) == len(X_list) * batch_size

# %% ../nbs/tests.ipynb 8
def test_dataloader(cls, ds_type='jax', samples=1000, batch_size=12):
    feats = np.arange(samples).repeat(10).reshape(samples, 10)
    labels = np.arange(samples).reshape(samples, 1)

    if ds_type == 'jax':
        ds = ArrayDataset(feats, labels)
    elif ds_type == 'torch':
        ds = torch.utils.data.TensorDataset(
            torch.from_numpy(feats), torch.from_numpy(labels))
    elif ds_type == 'tf':
        ds = tf.data.Dataset.from_tensor_slices((feats, labels))
    elif ds_type == "hf":
        ds = hf_datasets.Dataset.from_dict({"feats": feats, "labels": labels})
    else:
        raise ValueError(f"Unknown ds_type: {ds_type}")
    
    test_no_shuffle(cls, ds, batch_size, feats, labels)
    test_no_shuffle_drop_last(cls, ds, batch_size, feats, labels)
    test_shuffle(cls, ds, batch_size, feats, labels)
    test_shuffle_drop_last(cls, ds, batch_size, feats, labels)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/experimental/mp.ipynb.

# %% ../../nbs/experimental/mp.ipynb 1
from __future__ import print_function, division, annotations
from ..imports import *
from ..datasets import ArrayDataset, JAXDataset
from ..loaders import BaseDataLoader
from ..utils import get_config, asnumpy
from ..tests import *
import jax_dataloader as jdl
from threading import Thread, Event
from queue import Queue, Full
import multiprocessing as mp
import weakref

# %% auto 0
__all__ = ['chunk', 'EpochIterator']

# %% ../../nbs/experimental/mp.ipynb 2
def chunk(seq: Sequence, size: int) -> List[Sequence]:
    return [seq[pos:pos + size] for pos in range(0, len(seq), size)]  


# %% ../../nbs/experimental/mp.ipynb 3
class EpochIterator(Thread):
    """[WIP] Multiprocessing Epoch Iterator"""
    
    def __init__(self, data, batch_size: int, indices: Sequence[int]):
        super().__init__()
        self.data = data
        batches = chunk(indices, batch_size)
        self.iter_idx = iter(batches)
        self.output_queue = Queue(5) # TODO: maxsize
        self.terminate_event = Event()
        self.start()

    def run(self):
        try:
            while True:
                # get data
                result = self.get_data()
                # put result in queue
                while True:
                    try: 
                        self.output_queue.put(result, block=True, timeout=0.5)
                        break
                    except Full: pass
                
                    if self.terminate_event.is_set(): return                

        except StopIteration:
            self.output_queue.put(None)

    def __next__(self):
        result = self.output_queue.get()
        if result is None:
            self.close()
            raise StopIteration()
        return result
    
    def __iter__(self):
        return self
    
    def __del__(self):
        self.close()

    def close(self):
        self.terminate_event.set()

    def get_data(self):
        batch_idx = next(self.iter_idx)
        batch = self.data[batch_idx]
        return batch




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/loader.jax.ipynb.

# %% ../../nbs/loader.jax.ipynb 3
from __future__ import print_function, division, annotations
from ..imports import *
from ..datasets import ArrayDataset, JAXDataset
from . import BaseDataLoader
from ..utils import get_config, asnumpy
from ..tests import *
import jax_dataloader as jdl
from threading import Thread, Event
from queue import Queue

# %% auto 0
__all__ = ['EpochIterator', 'to_jax_dataset', 'DataLoaderJAX']

# %% ../../nbs/loader.jax.ipynb 4
def EpochIterator(
    data,
    batch_size: int,
    indices: Sequence[int]
):
    for i in range(0, len(indices), batch_size):
        idx = indices[i:i+batch_size]
        yield data[idx]

# %% ../../nbs/loader.jax.ipynb 5
@dispatch
def to_jax_dataset(dataset: JAXDataset):
    if isinstance(dataset, ArrayDataset):
        dataset.asnumpy()
    return dataset

@dispatch
def to_jax_dataset(dataset: HFDataset):
    return dataset.with_format('numpy')

# %% ../../nbs/loader.jax.ipynb 6
class DataLoaderJAX(BaseDataLoader):

    @typecheck
    def __init__(
        self, 
        dataset: Union[JAXDataset, HFDataset], 
        batch_size: int = 1,  # batch size
        shuffle: bool = False,  # if true, dataloader shuffles before sampling each batch
        num_workers: int = 0,  # how many subprocesses to use for data loading. Ignored.
        drop_last: bool = False,
        **kwargs
    ):
        self.key = jrand.PRNGKey(get_config().global_seed)
        self.dataset = to_jax_dataset(dataset)
        
        self.indices = np.arange(len(dataset))
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last
    
    def __iter__(self):
        # shuffle (permutation) indices every epoch        
        indices = jrand.permutation(self.next_key(), self.indices).__array__() if self.shuffle else self.indices
        
        if self.drop_last:
            indices = indices[:len(self.indices) - len(self.indices) % self.batch_size]
        return EpochIterator(self.dataset, self.batch_size, indices)
    
    def next_key(self):
        self.key, subkey = jrand.split(self.key)
        return subkey
    
    def __len__(self):
        return len(self.indices) // self.batch_size + int(not self.drop_last)
